"""Subpackage with Rba model classes
copied from rbaxdf and extended to create
RBA model from xdf model
Peter Schubert, 20.09.2023 (HHU, CCB)
"""

from .rba_model import RbaModel


__all__ = ['RbaModel']
